;;;; box.scm
;;;; Kon Lovett, Oct '08

;; Issues
;;
;; - All operations inlined & primitive due to high-performance nature.
;;
;; - Note that 'procedure-data' returns #f for anything other than an extended-procedure.

;;; Prelude

(declare
  (usual-integrations)
  (disable-interrupts)
  (fixnum)
  (inline)
  (local)
  (no-procedure-checks)
  (disable-warning redef) ;##sys#procedure->string is redefined!
  (bound-to-procedure
    ##sys#signal-hook
    ##sys#procedure->string ) )

(include "chicken-primitive-object-inlines")

;;

(define-inline (%->boolean obj) (and obj #t))


;;; Box Structure Support

(define-inline (%make-box tag init) (%make-structure tag init))

(define-inline (%box-structure-mutable? obj) (%structure-instance? obj 'box!))

(define-inline (%box-structure-immutable? obj) (%structure-instance? obj 'box))

(define-inline (%box-structure? obj)
  (and (or (%box-structure-mutable? obj) (%box-structure-immutable? obj))
       (%fx= 2 (%structure-length obj)) ) )

(define-inline (%box-structure-tag obj) (and (%box-structure? obj) (%structure-tag obj)))

(define-inline (%box-structure-ref box) (%structure-ref box 1))

(define-inline (%box-structure-set! box obj) (%structure-set! box 1 obj))

;;; Box Procedure Support

;; Box Variable

(define-inline (%box-variable-immutable-tag? obj) (%eq? 'boxvar obj))

(define-inline (%box-variable-mutable-tag? obj) (%eq? 'boxvar! obj))

(define-inline (%box-variable-tag? obj)
  (or (%box-variable-mutable-tag? obj) (%box-variable-immutable-tag? obj) ) )

(define-inline (%box-variable? obj)
  (and-let* ((dat (procedure-data obj)))
    (%box-variable-tag? dat) ) )

;; Box Location

(define-inline (%box-location-immutable-tag? obj) (%eq? 'boxloc obj))

(define-inline (%box-location-mutable-tag? obj) (%eq? 'boxloc! obj))

(define-inline (%box-location-tag? obj)
  (or (%box-location-mutable-tag? obj) (%box-location-immutable-tag? obj) ) )

(define-inline (%box-location? obj)
  (and-let* ((dat (procedure-data obj)))
    (%box-location-tag? dat) ) )

;; Box Procedure

(define-inline (%box-procedure-tag? obj) (or (%box-variable-tag? obj) (%box-location-tag? obj)))

(define-inline (%box-procedure-tag obj)
  (and-let* ((dat (procedure-data obj))
             ((%box-procedure-tag? dat)))
    dat ) )

(define-inline (%box-procedure? obj) (%->boolean (%box-procedure-tag obj)))

(define-inline (%box-procedure-immutable? obj)
  (and-let* ((dat (procedure-data obj)))
    (or (%box-variable-immutable-tag? dat) (%box-location-immutable-tag? dat) ) ) )

(define-inline (%box-procedure-mutable? obj)
  (and-let* ((dat (procedure-data obj)))
    (or (%box-variable-mutable-tag? dat) (%box-location-mutable-tag? dat) ) ) )

;; Box Procedure Operations

(define-inline (%box-procedure-ref box) (box (lambda (ref set loc) (ref))))

(define-inline (%box-procedure-set! box obj) (box (lambda (ref set loc) (set obj))))

(define-inline (%box-procedure-location box) (box (lambda (ref set loc) (loc))))

;;

(define-inline (%box? obj) (or (%box-structure? obj) (%box-procedure? obj)))

;; Print

(define-inline (%box-print box)
  (let ((val (cond ((%box-structure? box)  (%box-structure-ref box))
                   ((%box-procedure? box)  (%box-procedure-ref box))
                   (else
                    (error-box-type 'box-print box)))))
	  (display "#&") (write val)))


;;; Module box

(require-library ports lolevel)

(module box (;export
  make-box (make-box-variable $finvar) (make-box-location $finloc)
  box? box-variable? box-location?
  box-mutable? box-immutable?
  box-set! box-ref
  box-location
  box
  set-box! unbox
  ;
  *box-structure? *box-structure-ref *box-structure-set!
  *box-procedure? *box-procedure-ref *box-procedure-set!)

(import
  scheme
  (only chicken
    optional                ;due to #!optional implementation
    let-optionals           ;due to #!optional implementation
    define-reader-ctor
    define-record-printer
    let-location
    and-let*
    getter-with-setter
    void
    set-sharp-read-syntax!
    ##sys#signal-hook
    ##sys#procedure->string)
  (only ports
    with-output-to-port
    with-output-to-string)
  (only lolevel
    extend-procedure procedure-data
    make-weak-locative make-locative))


;;; Internals

;; Errors

(define (error-box-immutable loc box . args)
  (apply ##sys#signal-hook #:type-error loc "bad argument type - not a mutable box" box args))

(define (error-box-type loc obj . args)
  (apply ##sys#signal-hook #:type-error loc "bad argument type - not a box" obj args))

;; Finishers

(define ($finvar tag ref set)
  (extend-procedure (lambda (proc) (proc ref set (lambda () (location (ref))))) tag))

(define ($finloc tag ref set loc)
  (extend-procedure (lambda (proc) (proc ref set loc)) tag))


;;; Box

;; Direct calls
;; For use by high-performance routines (such as core routine replacements)

(define (*box-structure? obj) (%box-structure? obj))
(define (*box-structure-ref box) (%box-structure-ref box))
(define (*box-structure-set! box val) (%box-structure-set! box val))

(define (*box-procedure? obj) (%box-procedure? obj))
(define (*box-procedure-ref box) (%box-procedure-ref box))
(define (*box-procedure-set! box val) (%box-procedure-set! box val))

;; Constructers

(define-syntax make-box-variable
  (syntax-rules ()
    ((_ ?var)
     (make-box-variable ?var #f))
    ((_ ?var ?immutable?)
     #;(identifier? ?var)
     ($finvar
      (if ?immutable? 'boxvar 'boxvar!)
      (lambda () ?var)
      (if ?immutable? (void) (lambda (val) (set! ?var val)))))))

(define-syntax make-box-location
  (syntax-rules ()
    ((_ ?typ ?val)
     (make-box-location ?typ ?val #f))
    ((_ ?typ ?val ?immutable?)
     #;(identifier? ?typ)
     (let-location ((var ?typ ?val))
       ($finloc
        (if ?immutable? 'boxloc 'boxloc!)
        (lambda () var)
        (if ?immutable? (void) (lambda (val) (set! var val)))
        (lambda () (location var)))))))

(define (make-box #!optional init immutable?) (%make-box (if immutable? 'box 'box!) init))

;; Predicates

(define (box? obj) (%box? obj))

(define (box-variable? obj) (%box-variable? obj))

(define (box-location? obj) (%box-location? obj))

(define (box-immutable? obj) (or (%box-structure-immutable? obj) (%box-procedure-immutable? obj)))

(define (box-mutable? obj) (or (%box-structure-mutable? obj) (%box-procedure-mutable? obj)))

;; Mutators

(define (box-set! box val)
  (cond ((%box-structure-tag box) =>
         (lambda (tag)
           (case tag
             ((box!) (%box-structure-set! box val))
             ((box)  (error-box-immutable 'box-set! box val)))))
        ((%box-procedure-tag box) =>
         (lambda (tag)
           (case tag
             ((boxvar! boxloc!) (%box-procedure-set! box val))
             ((boxvar boxloc)   (error-box-immutable 'box-set! box val)))))
        (else
         (error-box-type 'box-set! box val))))

;; Assessors

(define box-ref
  (getter-with-setter
    (lambda (box)
      (cond ((%box-structure? box)  (%box-structure-ref box))
            ((%box-procedure? box)  (%box-procedure-ref box))
            (else                   (error-box-type 'box-ref box))))
    box-set!))

(define (box-location box #!optional (weak? #f))
  (cond ((%box-structure? box)   ((if weak? make-weak-locative make-locative) box 1))
        ((%box-procedure? box)   (box (lambda (ref set loc) (loc))))
        (else                    (error-box-type 'box-location box))))


;;; MZ Scheme Style

(define-syntax box
  (syntax-rules ()
    ((_ ?arg0 ...) (make-box ?arg0 ...))))

(define-syntax unbox
  (syntax-rules ()
    ((_ ?box) (box-ref ?box))))

(define-syntax set-box!
  (syntax-rules ()
    ((_ ?box ?val) (box-set! ?box ?val))))


;;; Read/Print Syntax

(set-sharp-read-syntax! #\& (lambda (p) (make-box (read p))))

(define-reader-ctor 'box make-box)

(define-record-printer (box x p) (with-output-to-port p (lambda () (%box-print x))))

(define-record-printer (box-immutable x p) (with-output-to-port p (lambda () (%box-print x))))

(set! ##sys#procedure->string
  (let ((##sys#procedure->string ##sys#procedure->string))
    (lambda (x)
			(if (%box? x) (with-output-to-string (lambda () (%box-print x)))
					(##sys#procedure->string x)))))

) ;module box
